#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair, Paul Rougieux and Viorel Blujdea.

JRC Biomass Project.
Unit D1 Bioeconomy.

"""

# Built-in modules #
from functools import cached_property
import re
import textwrap

# Third party modules #
import yaml, pandas
from p_tqdm import p_umap, t_map

# First party modules #
from autopaths import Path
from plumbing.timer import Timer

# Internal modules #
from eu_cbm_hat import eu_cbm_data_dir
from eu_cbm_hat.core.runner import Runner
from eu_cbm_hat.cbm.dynamic import DynamicRunner

# Constant directory for all the data #
yaml_dir = eu_cbm_data_dir + "combos/"


def run_one_country(args):
    """Run a single country, allowing any errors with a broad except statement.

    This function should only be used by a Combination.run() method.
    """
    last_year, runner = args
    runner.num_timesteps = last_year - runner.country.inventory_start_year
    print(runner)
    try:
        # The argument interrupt_on_error=False deals with errors happening
        # during the actual libcbm run
        runner.run(verbose=True, interrupt_on_error=False)
    # Catching general exception in case there are other errors in the input
    # data preparation or pre processor
    except Exception as general_error:
        print(general_error)


def runner_log_summary(runner, patterns=None):
    """Return the summary of a past model run for the given runner

    Usage:

        >>> from eu_cbm_hat.core.continent import continent
        >>> from eu_cbm_hat.combos.base_combo import runner_log_summary
        >>> runner_lu = continent.combos['reference'].runners['LU'][-1]
        >>> print(runner_log_summary(runner_lu))

    """
    if patterns is None:
        patterns = ["error", "done", "elapsed", "disk", "total elapsed"]

    msg = f"\n * {runner.country.iso2_code} {runner.country.country_name}\n"

    if not runner.paths.log.exists:
        msg += f"No log file at: {runner.paths.log}\n"
    else:
        msg += f"{runner.paths.log}\n"
        with open(runner.paths.log, "r") as file:
            for line in file:
                if any(pattern in line.lower() for pattern in patterns):
                    msg += f"{line}"

    return msg


class Combination(object):
    """
    This object represents a combination of specific scenarios for different
    activities and includes any other customization of a given model run.

    Each Combination subclass must define a list of Runner instances as
    the <self.runners> property. This enables the complete customization of
    any Runner by the specific Combination instance.

    You can run a combo like this:

        >>> from eu_cbm_hat.core.continent import continent
        >>> combo = continent.combos['historical']
        >>> combo()

    You can run a specific runner from a given country like this:

        >>> from eu_cbm_hat.core.continent import continent
        >>> combo = continent.combos['historical']
        >>> r = combo.runners['LU'][-1]
        >>> r.run(True, True, True)


    Print information on the inventory and yield curve used in the management
    activity for this runner

      print("Scenario combinations")
      for key in ["inventory", "growth_curves", "events"]:
          print(key, ":", runner.combo.config[key])

    You can then check the output pools:

        >>> r.output.load('pools')
    """

    def __init__(self, continent, short_name=None):
        self.short_name = short_name
        # Save parent #
        self.continent = continent
        # The combos dir used for all output #
        self.output_dir = self.continent.output_dir
        # The base dir for our output #
        self.base_dir = Path(self.output_dir + self.short_name + "/")
        # The path to our specific YAML file #
        self.yaml_path = yaml_dir + self.short_name + ".yaml"

    def __repr__(self):
        return "%s object with %i runners" % (self.__class__, len(self))

    def __iter__(self):
        return iter(self.runners.values())

    def __len__(self):
        return len(self.runners.values())

    def __getitem__(self, key):
        """Return a runner based on a country code."""
        return self.runners[key]

    @cached_property
    def config(self) -> dict:
        """
        The values chosen by the user in the YAML file which decide on every
        scenario choice for every activity and silvicultural practice.
        """
        # Read it with a third party library #
        with open(self.yaml_path, "r") as handle:
            result = yaml.safe_load(handle)
        # Convert silvicultural choices to dataframes #
        key = "harvest"
        value = result[key]
        if not isinstance(value, str):
            df = pandas.DataFrame.from_dict(value, orient="index", columns=["scenario"])
            df = df.rename_axis("year").reset_index()
            result[key] = df
        # Return result #
        return result

    @cached_property
    def runners(self) -> dict:
        """
        A dictionary of country codes as keys with a list of runners as
        values.
        """
        accepted_runner_types = ["base_runner", "dynamic_runner"]

        # Error message to be reused by diverse errors
        msg = f"The yaml file at {self.yaml_path} "
        msg += "should contain a 'runner_type' field with values:"
        msg += "\nrunner_type: "
        msg += "\n# or\nrunner_type: ".join(accepted_runner_types)

        # If runner_type is not defined in the yaml file raise an error
        if "runner_type" not in self.config.keys():
            raise ValueError(msg)

        if self.config["runner_type"] not in accepted_runner_types:
            msg_2 = f"runner_type: {self.config['runner_type']} "
            msg_2 += "is not an accepted value.\n"
            raise ValueError(msg_2 + msg)

        # If it's defined as "base_runner", return base runners
        if self.config["runner_type"] == "base_runner":
            return {c.iso2_code: [Runner(self, c, 0)] for c in self.continent}

        # If it's defined as "dynamic_runner" return dynamic runners
        if self.config["runner_type"] == "dynamic_runner":
            return {c.iso2_code: [DynamicRunner(self, c, 0)] for c in self.continent}

    def __call__(self, parallel=False, timer=True):
        """Legacy code, use the .run() method instead

        A method to run a combo by simulating all countries.

        Compared to the run() method, this call also runs many steps, if the
        runner has many steps inside .
        """
        # Message #
        print("Running combo '%s'." % self.short_name)
        # Timer start #
        timer = Timer()
        timer.print_start()

        # Function to run a single country #
        def run_country(args):
            code, steps = args
            for runner in steps:
                return runner.run()

        # Run countries sequentially #
        if not parallel:
            result = t_map(run_country, self.runners.items())
        # Run countries in parallel #
        else:
            result = p_umap(run_country, self.runners.items(), num_cpus=4)
        # Timer end #
        timer.print_end()
        timer.print_total_elapsed()
        # Compile logs #
        self.compile_logs()
        # Return #
        return result

    def run(self, last_year: int, countries: list = None, parallel=True):
        """Run a scenario combination

        If the list of countries is not specified, run all countries. A
        convenient method that makes it possible to run all countries inside a
        combination of scenarios. If one country fails to run, the error will
        be kept in its log files but the other countries will continue to run.

        Note: this method makes use of the run_one_country() function above
        which will only run one step inside the country. An update to that
        function will be needed in case your simulation needs many steps. We
        typically only run one step normally. Here the meaning of step is not
        that of yearly time steps, but bigger steps in terms of being able to
        start and stop the model which were foreseen in a legacy version of the
        model.

        Usage:

            >>> from eu_cbm_hat.core.continent import continent
            >>> # Run the selected list of countries
            >>> continent.combos["reference"].run(2050, ['IT','ZZ'])
            >>> # Run all countries with parallel cpus
            >>> continent.combos["reference"].run(2050)
            >>> # Run sequentially (not in parallel)
            >>> continent.combos["reference"].run(2050, parallel=False)

        """
        if countries is None:
            countries = self.runners.keys()
        # List of tuples, each tuple will be passed as argument to the
        # function that runs one country.
        runner_items = [(last_year, self.runners[k][-1]) for k in countries]
        if parallel:
            result = p_umap(run_one_country, runner_items, num_cpus=10)
        else:
            result = t_map(run_one_country, runner_items)
        return result

    def compile_logs(self, step=-1):
        # Open file #
        summary = self.base_dir + "all_logs.md"
        summary.open(mode="w")
        # Write title #
        title = "# Summary of all log files #\n\n"
        summary.handle.write(title)
        # Loop over runners #
        for rs in self.runners.values():
            r = rs[step]
            summary.handle.write("\n## " + r.country.country_name)
            summary.handle.write(" (" + r.country.iso2_code + ")" + "\n\n")
            content = textwrap.indent(r.paths.log.contents, "    ")
            summary.handle.write(content)
        # Close #
        summary.close()
        # Message #
        msg = "Log files compiled at:\n\n%s\n"
        print(msg % summary)
        # Return #
        return summary

    def print_log_summary(self, patterns=None, short=False):
        """Print the summary of all runners in a scenario combination

        Change the patterns argument to display more information from the log
        files.

        Usage:

            >>> from eu_cbm_hat.core.continent import continent
            >>> continent.combos["reference"].print_log_summary()

        For example, to check for a specific log message containing the words
        'carbon' and 'salvage', call:

            >>> continent.combos["reference"].print_log_summary(['carbon', 'salvage'])

        """
        if isinstance(patterns, str):
            patterns = [patterns]
        country_logs = {}

        for country_code in self.runners.keys():
            if country_code == "ZZ":
                continue
            runner = self.runners[country_code][-1]
            try:
                country_logs[country_code] = runner_log_summary(
                    runner, patterns=patterns
                )
            except OSError as e:  # Sometimes the log file is unreadable
                country_logs[country_code] = str(e)
            if patterns is not None and "done" not in patterns:
                continue
            # Summary one line per country at the beginning
            msg = ""
            if "done" in country_logs[country_code].lower():
                for line in country_logs[country_code].split("\n"):
                    if "total elapsed time" in line.lower():
                        msg += str(re.findall("(elapsed time: .*)", line)[-1])
                    if "done" in line.lower():
                        msg += " " + line
                print(re.sub("\n", " ", msg))
            else:
                print(self.short_name, country_code, "No 'done' message.")

        if short:
            return

        print("\n--------------------------------------------------------------")
        print("-------------------- Details ---------------------------------")
        print("--------------------------------------------------------------")

        for country_code, log in country_logs.items():
            print(country_code, log)



""" Post processing of the EU_CBM_HAT scenario combinations output

See post_processor/agg_combos.py for documentation of the save_agg_combo_output

Usage:

    cd $HOME/repos/eu_cbm/eu_cbm_hat/scripts/post_processing
    # or on BDAP
    cd $HOME/eu_cbm/eu_cbm_hat/scripts/post_processing
    ipython -i process_scenario_combo.py -- --combo_names reference
    ipython -i process_scenario_combo.py -- --combo_names pikssp2 pikfair
    ipython -i process_scenario_combo.py -- --combo_names pikssp2_fel1 pikssp2_owc_max pikssp2_owc_min
    ipython -i process_scenario_combo.py -- --combo_names pikfair_fel1 pikfair_owc_max pikfair_owc_min
    ipython -i process_scenario_combo.py -- --combo_names pikssp2_fel1 pikssp2_owc_max pikssp2_owc_min pikfair_fel1 pikfair_owc_max pikfair_owc_min

Share the aggregated output to the temporary transfer directory

    rsync -zav $HOME/eu_cbm/eu_cbm_data/output_agg /eos/jeodpp/data/projects/SUSBIOM-TRADE/transfer/eu_cbm_data

Separate processing to update a single data frame

    >>> from eu_cbm_hat.post_processor.agg_combos import apply_to_all_countries
    >>> from eu_cbm_hat.post_processor.agg_combos import nai_one_country
    >>> from eu_cbm_hat.post_processor.agg_combos import output_agg_dir
    >>> combo_name = "pikssp2_fel1"
    >>> combo_dir = output_agg_dir / combo_name
    >>> nai_st = apply_to_all_countries(nai_one_country, combo_name=combo_name, groupby=["status"])
    >>> nai_st.to_parquet(combo_dir / "nai_by_year_st_test_to_delete.parquet")
    >>> nai_sf = apply_to_all_countries(nai_by_sf_one_country, combo_name=combo)
    >>> nai_sf.to_parquet(combo_dir / "nai_by_sf.parquet")

"""

from p_tqdm import p_umap
import argparse
from eu_cbm_hat.post_processor.agg_combos import save_agg_combo_output

parser = argparse.ArgumentParser(
    description="Post processing of the EU_CBM_HAT scenario combinations"
)
parser.add_argument(
    "--combo_names", nargs="+", default=None, help="List of names of scenario combos"
)

shell_args = parser.parse_args()
COMBO_NAMES = shell_args.combo_names

p_umap(save_agg_combo_output, COMBO_NAMES, num_cpus=6)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.

Usage:

    >>> from eu_cbm_hat.core.continent import continent
    >>> runner = continent.combos['special'].runners["ZZ"][-1]
    >>> runner.qaqc.aidb_check.run_all_checks()

"""


# Third party modules #
import numpy
import pandas


def check_dist_matrix_prop(df: pandas.DataFrame):
    """Check if the disturbance matrix proportions sum to one

    Usage:

        >>> from eu_cbm_hat.core.continent import continent
        >>> from eu_cbm_hat.qaqc.aidb import check_dist_matrix_prop
        >>> runner = continent.combos['reference'].runners['AT'][-1]
        >>> dist_matrix_value = runner.country.aidb.db.read_df('disturbance_matrix_value')
        >>> check_dist_matrix_prop(dist_matrix_value)

    """
    id_cols = ["disturbance_matrix_id", "source_pool_id"]
    df_check = (
        df.groupby(id_cols)["proportion"]
        .agg("sum")
        .reset_index()
        .rename(columns={"proportion": "proportion_sum"})
    )
    prop_issue = df_check.query("proportion_sum < 0.99 or proportion_sum > 1.001")
    if not prop_issue.empty:
        msg = "The disturbance matrix proportion don't sum to one."
        msg += "\nSummary data frame:"
        msg += f"\n{prop_issue}"
        msg += f"\n{prop_issue[id_cols].merge(df, on=id_cols)}"
        raise ValueError(msg)



class AIDBCheck:
    """
    Check the consistency of the AIDB

        >>> from eu_cbm_hat.core.continent import continent
        >>> runner = continent.combos['reference'].runners["ZZ"][-1]

    Check id duplication in vol_to_bio_factor

        >>> runner.qaqc.aidb_check.check_vol_to_bio_factor_id_duplication()

    Check disturbance matrix proportion sum to one

        >>> runner.qaqc.aidb_check.check_dist_matrix_proportions_sum_to_one()
    """

    def __init__(self, qaqc):
        # Default attributes #
        self.runner = qaqc.runner
        self.aidb = self.runner.country.aidb

    def run_all_checks(self):
        """Run all AIDB checks"""
        self.check_vol_to_bio_factor_id_duplication()

    def check_vol_to_bio_factor_id_duplication(self):
        """Investigate identifier duplication in the vol_to_bio_factor in the AIDB"""
        df = self.aidb.db.read_df("vol_to_bio_factor")
        # Select duplicated rows
        selector = df["id"].duplicated(keep=False)
        if any(selector):
            msg = "Duplicated ids in the vol_to_bio_factor table in the AIDB."
            msg += "The following rows are duplicated:\n"
            msg += f"{df[selector]}\n\n"
            msg += (
                "Investigate by loading the table with a command similar to this one:\n"
            )
            msg += "runner.country.aidb.db.read_df('vol_to_bio_factor')"
            raise ValueError(msg)

    def check_dist_matrix_proportions_sum_to_one(self):
        """Check if the disturbance matrix proportions sum to one"""
        dist_matrix_value = self.aidb.db.read_df("disturbance_matrix_value")
        check_dist_matrix_prop(dist_matrix_value)



"""Run the Reference scenario combo in V2 and V2

Using the symbolic link defined in scripts/running/run_on_bdap.sh, make sure
that the repository switch works correctly before running the second scenario.

    cd $HOME
    ipython ~/eu_cbm/eu_cbm_hat/scripts/conversion/libcbmv2/switch_git_repos.py -- --version 1
    ./run.sh reference.py
    ipython ~/eu_cbm/eu_cbm_hat/scripts/conversion/libcbmv2/switch_git_repos.py -- --version 2
    ./run.sh refence_v2.py

Run only one country

    ipython ~/eu_cbm/eu_cbm_hat/scripts/conversion/libcbmv2/switch_git_repos.py -- --version 1
    ipython ~/eu_cbm/eu_cbm_hat/scripts/running/run_lu.py
    ipython ~/eu_cbm/eu_cbm_hat/scripts/conversion/libcbmv2/switch_git_repos.py -- --version 2
    ipython ~/eu_cbm/eu_cbm_hat/scripts/running/run_lu.py
    # Try another scenario
    cd $HOME/eu_cbm/eu_cbm_hat/scripts/running/
    ipython run_scenario_combo.py -- --combo_name pikssp2_fel1 --last_year 2070 --countries LU

Run the reference2 scenario with libcbm version 2 for a list of countries

    ipython ~/eu_cbm/eu_cbm_hat/scripts/conversion/libcbmv2/switch_git_repos.py -- --version 2
    cd $HOME/eu_cbm/eu_cbm_hat/scripts/running/
    ipython run_scenario_combo.py -- --combo_name reference2 --last_year 2070 --countries AT CZ DE DK EE ES FI IE IT LT LU LV NL RO SI SK ZZ
    ipython run_scenario_combo.py -- --combo_name reference2 --last_year 2070 --countries IT

"""




"""
The purpose of this script is to compute the Net Annual Increment for one country
"""

from functools import cached_property
from typing import Union, List
import warnings
import pandas

import numpy as np

from eu_cbm_hat.post_processor.convert import ton_carbon_to_m3_ob

POOLS_COLS = ["merch_stock_vol", "agb_stock_vol"]
FLUXES_COLS = [
    "merch_prod_vol",
    "other_prod_vol",
    "turnover_merch_input_vol",
    "turnover_oth_input_vol",
    "dist_merch_input_vol",
    "dist_oth_input_vol",
    "merch_air_vol",
    "oth_air_vol",
]
NAI_AGG_COLS = ["area"] + POOLS_COLS + FLUXES_COLS


def compute_nai_gai(df: pandas.DataFrame, groupby: Union[List[str], str]):
    """Compute the Net Annual Increment and Gross Annual Increment

    Based on stock change and movements to the product pools as well as
    turnover and mouvements to air.

    Before using this function, make sure you aggregate along the groupby
    variables and year first. Then the year should not be present on the
    groupby variable. For example:

        >>> from eu_cbm_hat.core.continent import continent
        >>> from eu_cbm_hat.post_processor.nai import NAI_AGG_COLS
        >>> from eu_cbm_hat.post_processor.nai import compute_nai_gai
        >>> runner = continent.combos['reference'].runners['ZZ'][-1]
        >>> index = ["status"]
        >>> nai_st = runner.post_processor.nai.df_agg(index)
        >>> nai_st_2  = nai_st.groupby(["year"] + index, observed=True)[NAI_AGG_COLS].agg("sum").reset_index()
        >>> nai_st_2 = compute_nai_gai(nai_st_2, groupby=index)
        >>> cols = ["year", "status", "nai_merch", "nai_agb"]
        >>> nai_st_2.query("status == 'ForAWS'")[cols].tail().round()
            year  status  nai_merch   nai_agb
        35  2025  ForAWS   435110.0  562391.0
        38  2026  ForAWS   428422.0  552260.0
        41  2027  ForAWS   421882.0  538261.0
        44  2028  ForAWS   422830.0  521565.0
        47  2029  ForAWS   414102.0  505772.0

    """
    if isinstance(groupby, str):
        groupby = [groupby]

    if "year" in groupby:
        msg = " This functions computes the difference in stock across groups "
        msg += "through time so 'year' should not be in the group by variables:\n"
        msg += f"{groupby}"
        raise ValueError(msg)

    # Order by groupby variables, then years
    df.sort_values(groupby + ["year"], inplace=True)

    # Check that there are no duplications over the groupby variables plus year
    selector = df[["year"] + groupby].duplicated(keep=False)
    if any(selector):
        msg = "The following rows have duplications along the groupby variables.\n"
        msg += f"{df.loc[selector, ['year'] + groupby ]}"
        msg += "\nPlease aggregate first along the groupby variables and year:\n"
        msg += f"{['year'] + groupby }\n Then run this function.\n"
        raise ValueError(msg)

    # Compute the difference in stock for the standing biomass
    # Use Observed = True to avoid the warning when using categorical variables
    df["net_merch"] = df.groupby(groupby, observed=True)["merch_stock_vol"].diff()
    df["net_agb"] = df.groupby(groupby, observed=True)["agb_stock_vol"].diff()

    # Compute NAI for the merchantable pool
    df["nai_merch"] = df[["net_merch", "merch_prod_vol", "dist_merch_input_vol"]].sum(
        axis=1
    )
    df["gai_merch"] = df["nai_merch"] + df[
        ["turnover_merch_input_vol", "merch_air_vol"]
    ].sum(axis=1)

    # Compute NAI for the merchantable pool and OWC pool together
    df["nai_agb"] = df[
        [
            "net_agb",
            "merch_prod_vol",
            "other_prod_vol",
            "dist_merch_input_vol",
            "dist_oth_input_vol",
        ]
    ].sum(axis=1)
    df["gai_agb"] = df["nai_agb"] + df[
        [
            "turnover_merch_input_vol",
            "turnover_oth_input_vol",
            "merch_air_vol",
            "oth_air_vol",
        ]
    ].sum(axis=1)
    # Compute per hectare values
    df["nai_merch_ha"] = df["nai_merch"] / df["area"]
    df["gai_merch_ha"] = df["gai_merch"] / df["area"]
    df["nai_agb_ha"] = df["nai_agb"] / df["area"]
    df["gai_agb_ha"] = df["gai_agb"] / df["area"]
    return df


class NAI:
    """Compute the net annual increment$

    Usage:

        >>> from eu_cbm_hat.core.continent import continent
        >>> runner = continent.combos['reference'].runners['ZZ'][-1]
        >>> pfv = runner.post_processor.nai.pools_fluxes_vol
        >>> pfv[["year", "disturbance_type", "merch_stock_vol", "agb_stock_vol"]].round()
             year  disturbance_type  merch_stock_vol  agb_stock_vol
        0    1999                 0         210788.0       288140.0
        1    1999                 0         260786.0       356455.0
        2    2000                 0         220762.0       300918.0
        3    2000                 0         273105.0       372236.0
        4    2001                 0         219576.0       299263.0
        ..    ...               ...              ...            ...
        855  2028                 0         648940.0       881020.0
        856  2029                 0         958579.0      1613398.0
        857  2029                 0         174177.0       292668.0
        858  2029                 0        3615729.0      4908137.0
        859  2029                 0         653509.0       887099.0
        <BLANKLINE>
        [860 rows x 4 columns]

        >>> # Net Annual Increment of the merchantable pool (nai_merch) and of
        >>> # all the above ground biomass (nai_agb) by status
        >>> nai_st = runner.post_processor.nai.df_agg(["status"])
        >>> selector = nai_st["status"] == 'ForAWS'
        >>> nai_st.loc[selector, ["year", "status", "area", "nai_merch", "nai_agb"]].head()
           year  status         area      nai_merch        nai_agb
        0  1999  ForAWS  91880.00000       0.000000       0.000000
        1  2000  ForAWS  91880.00000  362883.178649  469452.158547
        2  2001  ForAWS  91879.99912  362494.285355 -488554.806845
        3  2002  ForAWS  91879.99877  378072.436366  212918.717178
        4  2003  ForAWS  91879.99841  397329.021290  496215.446031

        >>> # TODO fix this example
        >>> # NAI per ha by status and forest type at country level
        >>> # See message below on KeyError: 'forest_type'
        >>> # when merging back the movements of the NF pools to products.
        >>> runner.post_processor.nai.df_agg(["status", "forest_type"]) # doctest: +SKIP

        >>> df = runner.post_processor.nai.df_agg(["status"])
        >>> df["nai_merch"] = df["nai_merch_ha"] * df["area"]
        >>> df_st = df.groupby(["year", "status"])[["area", "nai_merch"]].agg("sum").reset_index()
        >>> df_st["nai_merch_ha"] = df_st["nai_merch"] / df_st["area"]
        >>> df_st.tail()
            year  status          area      nai_merch  nai_merch_ha
        44  2028  ForAWS  86236.998804  422829.931377      4.903115
        45  2028      NF   5643.000250   28488.860714      5.048531
        46  2029      AR     90.000000      16.971735      0.188575
        47  2029  ForAWS  85727.998714  414101.528162      4.830412
        48  2029      NF   6152.000250    3272.171429      0.531887

        >>> # Plot NAI per ha by status (remove NF)
        >>> df_st = runner.post_processor.nai.df_agg(["status"])
        >>> selector = df_st["status"] != 'NF'
        >>> df_st = df_st.loc[selector].pivot(columns="status", index="year", values="nai_merch_ha")
        >>> from matplotlib import pyplot as plt
        >>> df_st.plot(ylabel="nai_merch m3 / ha") # doctest: +SKIP
        >>> plt.show()  # doctest: +SKIP

        Note in terms of ecosystem indicators, foliage should be there in the NAI
        increment computation.
        """

    def __init__(self, parent):
        self.parent = parent
        self.runner = parent.runner
        self.combo_name = self.runner.combo.short_name

    @cached_property
    def pools_fluxes_vol(self):
        """Merchantable pools and fluxes aggregated at the classifiers level"""
        df = self.parent.pools_fluxes_morf
        # Add wood density information by forest type
        df = df.merge(self.parent.wood_density_bark_frac, on="forest_type")

        # Convert tons of carbon to volume over bark
        df["merch_stock_vol"] = ton_carbon_to_m3_ob(df, "merch")
        df["agb"] = df["merch"] + df["other"]
        df["agb_stock_vol"] = ton_carbon_to_m3_ob(df, "agb")
    
        # Fluxes to products
        df["merch_prod_vol"] = ton_carbon_to_m3_ob(df, "merch_prod")
        df["other_prod_vol"] = ton_carbon_to_m3_ob(df, "oth_prod")

        # Fluxes which represent the biomass lost to the air
        df["merch_air_vol"] = ton_carbon_to_m3_ob(df, "disturbance_merch_to_air")
        df["oth_air_vol"] = ton_carbon_to_m3_ob(df, "disturbance_oth_to_air")

        df["turnover_merch_input_vol"] = ton_carbon_to_m3_ob(
            df, "turnover_merch_litter_input"
        )
        df["turnover_oth_input_vol"] = ton_carbon_to_m3_ob(
            df, "turnover_oth_litter_input"
        )

        # these filters for "== 0" are not needed as such transfers are zero anyway
        df["dist_merch_input_vol"] = np.where(
            df["disturbance_type"] == 0,
            0,
            ton_carbon_to_m3_ob(df, "disturbance_merch_litter_input"),
        )
        df["dist_oth_input_vol"] = np.where(
            df["disturbance_type"] == 0,
            0,
            ton_carbon_to_m3_ob(df, "disturbance_oth_litter_input"),
        )
        return df

    def df_agg(self, groupby: Union[List[str], str]):
        """Net Annual Increment aggregated by status and forest type

        Usage:

            >>> from eu_cbm_hat.core.continent import continent
            >>> runner = continent.combos['reference'].runners['ZZ'][-1]
            >>> # Net Annual Increment of the merchantable pool (nai_merch) and of
            >>> # all the above ground biomass (nai_agb) by status
            >>> nai_st = runner.post_processor.nai.df_agg(["status"])
            >>> selector = nai_st["status"] == 'ForAWS'
            >>> nai_st.loc[selector, ["year", "status", "area", "nai_merch", "nai_agb"]].head()
               year  status         area      nai_merch        nai_agb
            0  1999  ForAWS  91880.00000       0.000000       0.000000
            1  2000  ForAWS  91880.00000  362883.178649  469452.158547
            2  2001  ForAWS  91879.99912  362494.285355 -488554.806845
            3  2002  ForAWS  91879.99877  378072.436366  212918.717178
            4  2003  ForAWS  91879.99841  397329.021290  496215.446031

        """
        if isinstance(groupby, str):
            groupby = [groupby]
        if groupby != ["status"]:
            warnings.warn("This method was written for a group by status.")
        df = self.pools_fluxes_vol

        # Aggregate the sum of selected columns
        df_agg = (
            df.groupby(["year"] + groupby)[NAI_AGG_COLS].agg("sum").reset_index()
        )

        # Add NF movements to products back to ForAWS
        # Note this is a problem when we use grouping variables other than
        # "status" alone For example if groupby = ["status", "forest_type"]
        # the merge below generates the KeyError: 'forest_type'
        selector = df_agg["status"] == "NF"
        df_agg_nf = df_agg.loc[selector, ["year", "status"] + FLUXES_COLS].copy()
        df_agg_nf["status"] = "ForAWS"
        df_agg_nf.columns = df_agg_nf.columns.str.replace("_vol", "_vol_nf")
        df_agg = df_agg.merge(df_agg_nf, on=["year"] + groupby, how="left")
        fluxes_cols_nf = [x + "_nf" for x in FLUXES_COLS]
        df_agg[fluxes_cols_nf] = df_agg[fluxes_cols_nf].fillna(0)
        # Add the nf fluxes to the fluxes in ForAWS
        for col1, col2 in zip(FLUXES_COLS, fluxes_cols_nf):
            df_agg[col1] += df_agg[col2]

        # Compute NAI and GAI
        df_out = compute_nai_gai(df_agg, groupby=groupby)
        return df_out


    def df_agg_con_broad(self, groupby: Union[List[str], str]):
        """Net Annual Increment aggregated by status and con_broad. It WILL NOT work properly when there are transitions from con to broad and viceversa.
        Usage:
            >>> from eu_cbm_hat.core.continent import continent
            >>> runner = continent.combos['reference'].runners['ZZ'][-1]
            >>> # Net Annual Increment of the merchantable pool (nai_merch) and of
            >>> # all the above ground biomass (nai_agb) by status
            >>> nai_st = runner.post_processor.nai.df_agg_con_broad(["status", "con_broad"])
       
        """
        if isinstance(groupby, str):
            groupby = [groupby]
        if groupby != ["status", "con_broad"]:
            warnings.warn("This method was written for a group by status.")
        df = self.pools_fluxes_vol

        # Aggregate the sum of selected columns
        df_agg = (
            df.groupby(["year"] + groupby)[NAI_AGG_COLS].agg("sum").reset_index()
        )
        # Add NF movements to products back to ForAWS
        # Note this is a problem when we use grouping variables other than
        # "status" alone For example if groupby = ["status", "forest_type"]
        # the merge below generates the KeyError: 'forest_type'
        selector = df_agg["status"] == "NF"
        df_agg_nf = df_agg.loc[selector, ["year", "status", "con_broad"] + FLUXES_COLS].copy()
        df_agg_nf["status"] = "ForAWS"
        df_agg_nf.columns = df_agg_nf.columns.str.replace("_vol", "_vol_nf")
        df_agg = df_agg.merge(df_agg_nf, on=["year", "con_broad"] + groupby, how="left")
        fluxes_cols_nf = [x + "_nf" for x in FLUXES_COLS]
        df_agg[fluxes_cols_nf] = df_agg[fluxes_cols_nf].fillna(0)
        # Add the nf fluxes to the fluxes in ForAWS
        for col1, col2 in zip(FLUXES_COLS, fluxes_cols_nf):
            df_agg[col1] += df_agg[col2]
        # Compute NAI and GAI
        df_out_con_broad = compute_nai_gai(df_agg, groupby=groupby)
        df_out_con_broad = df_out_con_broad[df_out_con_broad["status"] != "NF"]
        return df_out_con_broad



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair, Paul Rougieux and Viorel Blujdea.

JRC Biomass Project.
Unit D1 Bioeconomy.

A list of all combo classes. A combo is a scenario combination.

Creating a new scenario combination only requires an update to the input files
and to a yaml file that specifies which scenarios will be used in the combo for
each input file:

- In `eu_cbm_data`

    - Add a  new yaml file in `eu_cbm_data/combos` to specify which scenario to
    use for each input file. A yaml file is a configuration file that can be
    edited with a text editor.

    - Add data specific to the “new_scenario” in each relevant input file. The
    assumptions have to be identified as “new_scenario” on the column
    “scenario” across all files.

        - For example, in case of a new scenario for afforestation, specific
        records must be added in: events.csv, growth_curves.csv, inventory.csv,
        transitions.csv. in

            ...\eu_cbm_data\countries\ZZ\activities\afforestation.

        - Also, information specific to the new assumptions must be added in
        the following files contained in the other directory (as explained
        above in Lower level of the EU-CBM-HAT: country specific inputs):

            ...\eu_cbm_data\countries\ZZ\common - on the new disturbance
            types;
            ...\eu_cbm_data\countries\ZZ\config - for the changes in the AIDB
            and association file;
            ...\eu_cbm_data\countries\ZZ\silv – description of the silvicultural
            practices and wood use.

        - Add data on IRW and FW demands in the new directory. irw_demand.csv,
        fw_demands.csv, rw_demands.csv in the directory
        ...\eu_cbm_data\demand\new_scenario.


[Old instructions before July 2023] Creating a new scenario combination in this
hard coded way is not required any more. The following instructions might still
be useful in case you need special python code in a scenario class that
modifies the behaviour of the scenario combinations beyond what the base runner
or the dynamic runner can do. In that case, and in that case only, you need to
change the source code of `eu_cbm_hat` in the following files:

    - Create a new file in ...\eu_cbm_hat\combos\new_scenario.py. Create a new
    class and import the corresponding module in the file __init__.py in
    ...\eu_cbm_hat\combos of the eu_cbm_hat. Add the new class
    [...new_scenario...] 

    - and import the new module [...new_...] in __init__.py, by paying
    attention to consistency of names, e.g., class name: “NewScenario”,
    short_name: “new_scenario”.

"""

# Built-in modules #

# First party modules #

# Internal modules #
from eu_cbm_hat.combos.reference         import Reference
from eu_cbm_hat.combos.ia_2040 import IA_2040

from eu_cbm_hat import eu_cbm_data_pathlib

# Hard coded combo classes
combo_classes = [Reference,
                 IA_2040,
                ]
combo_classes_dict = dict(zip([s.short_name for s in combo_classes], combo_classes))




"""Get harvest expected and provided"""

from typing import Union, List
from functools import cached_property
import numpy as np
import pandas
import yaml
from eu_cbm_hat.info.harvest import combined
from eu_cbm_hat.post_processor.convert import ton_carbon_to_m3_ub
from eu_cbm_hat.post_processor.convert import ton_carbon_to_m3_ob

"""
This dictionary is addedd to allow splitting the outputs on silvicultural practices. 
The correspondence of disturbance_types and silv_practices should be validf for all countries for the calibration period
"""
dist_silv_corresp = { 
                1 :'thinnings',#generic 5%
                1 :'thinnings',#generic 5% (calibration)
                2 :'salvage',#Wildfire
                3 :'final_cut',#Clearcut harvesting without salvage
                7 :'salvage',#Deforestation
                8 :'NA',#Afforestation
                10 :'thinnings',#10% commercial thinning
                11 :'thinnings',#generic 10%
                12 :'thinnings',#10% commercial thinning
                12 :'thinnings',#15% commercial thinning
                13 :'thinnings',#generic 15%
                13 :'thinnings',#generic 20%
                14 :'thinnings',#15% commercial thinning
                14 :'thinnings',#20% commercial thinning
                15 :'thinnings',#generic 20%
                15 :'thinnings',#generic 25%
                16 :'thinnings',#20% commercial thinning
                16 :'thinnings',#25% commercial thinning
                16 :'thinnings',#30% commercial thinning
                17 :'thinnings',#generic 25%
                17 :'thinnings',#generic 30%
                18 :'thinnings',#25% commercial thinning
                18 :'thinnings',#30% commercial thinning
                18 :'thinnings',#35% commercial thinning
                18 :'thinnings',#35% Commercial thinning
                19 :'thinnings',#35% commercial thinning
                21 :'final_cut',#97% clearcut
                22 :'final_cut',#Clearcut harvesting with salvage
                24 :'final_cut',#Clearcut with slash-burn
                40 :'thinnings',#generic 15% (calibration)
                40 :'final_cut',#Stand Replacing Natural Succession
                40 :'final_cut',#Stand Replacing Natural Succession (calibration)
                41 :'salvage',#generic 40% mortality (calibration)
                41 :'final_cut',#generic 90% mortality
                41 :'final_cut',#generic 90% mortality (calibration)
                41 :'salvage',#Insects with salvage logging
                41 :'salvage',#Insects with salvage logging (calibration)
                42 :'salvage',#generic 90% mortality (calibration)
                42 :'salvage',#Insects with salvage logging (calibration)
                42 :'salvage',#Insects with salvage logging (nd_nsr), Matrix ID 25
                43 :'salvage',#generic 60% mortality (calibration)
                43 :'salvage',#Salvage logging after insects (calibration)
                45 :'salvage',#generic 90% mortality (calibration)
                45 :'salvage',#Salvage logging after insects (calibration)
                50 :'salvage',#Fire with salvage logging
                50 :'salvage',#Fire with salvage logging (calibration)
                50 :'salvage',#generic 50% mortality (calibration)
                51 :'salvage',#Fire with salvage logging (calibration)
                115 :'thinnings',#15% commercial thinning
                120 :'thinnings',#generic 40% mortality
                125 :'final_cut',#generic 70%
                130 :'final_cut',#generic 85%
                400 :'final_cut',#Stand Replacing Natural Succession (projection)
                401 :'thinnings',#generic 15% (projection)
                401 :'final_cut',#generic 90% mortality (projection)
                401 :'final_cut',#Stand Replacing Natural Succession (projection)
                402 :'salvage',#Insects with salvage logging (projection)
                411 :'thinnings',#generic 40% mortality (projection)
                411 :'salvage',#generic 90% mortality (projection)
                411 :'salvage',#Insects with salvage logging (projection)
                420 :'salvage',#Insects with salvage logging (projection)
                421 :'salvage',#generic 90% mortality (projection)
                421 :'salvage',#Insects with salvage logging (projection)
                431 :'salvage',#generic 60% mortality (projection)
                431 :'salvage',#Salvage logging after insects (projection)
                451 :'salvage',#generic 90% mortality (projection)
                451 :'salvage',#Salvage logging after insects (projection)
                491 :'final_cut',#Stand Replacing Natural Succession (projection)
                500 :'salvage',#Fire with salvage logging (projection)
                501 :'salvage',#Fire with salvage logging (projection)
                501 :'salvage',#generic 50% mortality (projection)
                515 :'thinnings',#Post_conversion_LA_15%_commercial_thinning
                516 :'final_cut',#Conversion_to_u_u_con
                517 :'final_cut',#Conversion_to_u_u_broad
                518 :'final_cut',#Conversion_to_u_u_con
                535 :'thinnings',#Step_1_conversion_LA_35%_commercial_thinning
                550 :'thinnings',#Step_2_conversion_LA_50%_commercial_thinning
                615 :'thinnings',#Post_conversion_ST_15%_commercial_thinning
                625 :'thinnings',#Step_1_conversion_ST_25%_commercial_thinning
                640 :'thinnings',#Step_2_conversion_ST_40%_commercial_thinning
                700 :'final_cut',#Conversion_of_coppice_to_high_stands
                701 :'final_cut',#Conversion_of_old_stands_to_coppice
                1010 :'thinnings',#10% commercial thinning hist
                1111 :'thinnings',#generic 10% hist
                1212 :'thinnings',#10% commercial thinning hist
                1212 :'thinnings',#15% commercial thinning hist
                1313 :'thinnings',#generic 15% hist
                1313 :'thinnings',#generic 20% hist
                1414 :'thinnings',#15% commercial thinning hist
                1414 :'thinnings',#20% commercial thinning hist
                1515 :'thinnings',#generic 20% hist
                1515 :'thinnings',#generic 25% hist
                1616 :'thinnings',#20% commercial thinning hist
                1616 :'thinnings',#25% commercial thinning hist
                1616 :'thinnings',#30% commercial thinning hist
                1717 :'thinnings',#generic 25% hist
                1717 :'thinnings',#generic 30% hist
                1818 :'thinnings',#25% commercial thinning hist
                1818 :'thinnings',#30% commercial thinning hist
                1818 :'thinnings',#35% commercial thinning hist
                1818 :'thinnings',#35% Commercial thinning hist
                2121 :'final_cut',#97% clearcut hist
                2222 :'final_cut',#Clearcut harvesting with salvage hist
                2424 :'final_cut',#Clearcut with slash-burn hist
                4040 :'final_cut',#Stand Replacing Natural Succession (calibration) hist
                4141 :'thinnings',#generic 40% mortality (calibration) hist
                4141 :'salvage',#generic 90% mortality (calibration) hist
                4141 :'salvage',#generic 90% mortality hist
                4141 :'salvage',#Insects with salvage logging (calibration) hist
                4242 :'salvage',#generic 90% mortality (calibration) hist
                4242 :'salvage',#Insects with salvage logging (calibration) hist
                4343 :'salvage',#generic 60% mortality (calibration) hist
                4343 :'salvage',#Salvage logging after insects (calibration) hist
                4545 :'salvage',#generic 90% mortality (calibration) hist
                4545 :'salvage',#Salvage logging after insects (calibration) hist
                115115 :'thinnings',#15% commercial thinning hist
                120120 :'thinnings',#generic 40% mortality hist
                125125 :'salvage',#generic 70% hist
                130130 :'salvage',#generic 85% hist
                }

class Harvest:
    """Compute the harvest expected and provided

    Methods to load intermediate data frames used in the computation of the
    harvest expected and provided:

        >>> from eu_cbm_hat.core.continent import continent
        >>> runner = continent.combos['reference'].runners['LU'][-1]
        >>> runner.post_processor.harvest.demand
        >>> runner.post_processor.harvest.hat_events
        >>> runner.post_processor.harvest.hat_extra
        >>> runner.post_processor.harvest.expected_agg("year")
        >>> runner.post_processor.harvest.provided
        >>> runner.post_processor.harvest.provided_agg("year")
        >>> harvest_group = ["year", "forest_type", "mgmt_type", "mgmt_strategy", "con_broad", "disturbance_type"]
        >>> runner.post_processor.harvest.provided_agg(harvest_group)

        >>> runner.post_processor.harvest.expected_provided("year")
        >>> runner.post_processor.harvest.expected_provided(["year", "forest_type"])
        >>> runner.post_processor.harvest.expected_provided(["year", "disturbance_type"])
        >>> runner.post_processor.harvest.disturbance_types
        >>> runner.post_processor.harvest.area
        >>> runner.post_processor.harvest.area_agg(["year", "disturbance"])
        >>> runner.post_processor.harvest.area_agg(["year", "disturbance", "disturbance_type"])

    Plot harvest area by disturbance type through time

        >>> from matplotlib import pyplot as plt
        >>> area_agg = runner.post_processor.harvest.area_agg(["year", "disturbance"])
        >>> area_by_dist = area_agg.pivot(columns="disturbance", index="year", values="area")
        >>> area_by_dist.plot(title="LU harvest area by disturbance type")
        >>> plt.show()

    Plot harvest volume by disturbance type through time

        >>> harvest_prov_by_dist = area_agg.pivot(columns="disturbance", index="year", values="harvest_prov_ub")
        >>> harvest_prov_by_dist.plot(title="LU harvest volume by disturbance type")
        >>> plt.show()

    """

    def __init__(self, parent):
        self.parent = parent
        self.runner = parent.runner
        self.combo_name = self.runner.combo.short_name
        self.pools = self.parent.pools
        self.fluxes = self.parent.fluxes

    def __repr__(self):
        return '%s object code "%s"' % (self.__class__, self.runner.short_name)

    @cached_property
    def demand(self) -> pandas.DataFrame:
        """Get demand from the economic model using eu_cbm_hat/info/harvest.py

        Convert demand volumes from 1000m3 ub to m3 ub.
        """
        harvest_scenario_name = self.runner.combo.config["harvest"]
        irw = combined["irw"]
        irw["product"] = "irw_demand"
        fw = combined["fw"]
        fw["product"] = "fw_demand"
        df = pandas.concat([irw, fw]).reset_index(drop=True)
        # Convert volumes from 1000m3 ub to m3 ub
        df["value"] *= 1e3
        index = ["scenario", "iso2_code", "year"]
        df = df.pivot(index=index, columns="product", values="value").reset_index()
        df["rw_demand"] = df["fw_demand"] + df["irw_demand"]
        df = df.rename_axis(columns=None)
        selector = df["scenario"] == harvest_scenario_name
        selector &= df["iso2_code"] == self.runner.country.iso2_code
        return df[selector]

    @cached_property
    def hat_events(self) -> pandas.DataFrame:
        """Events from the harvest allocation tool

        Load HAT events which were saved in this line of cbm/dynamic.py:

            >>> self.runner.output.events = pandas.concat([self.runner.output.events, df[cols]])

        """
        # Load output events from the harvest allocation tool, generated in cbm/dynamic.py
        df = self.runner.output["events"]
        # Rename the amount expected by the Harvest Allocation Tool
        df.rename(columns={"amount": "amount_exp_hat"}, inplace=True)
        df["harvest_exp_hat"] = ton_carbon_to_m3_ub(df, "amount_exp_hat")
        # Check that the amount converted from tons of carbon back to cubic
        # meter gives the same value as the sum of irw_need and fw_colat
        for col in ["harvest_exp_hat", "irw_need", "fw_colat", "fw_need"]:
            df[col] = df[col].fillna(0)
        pandas.testing.assert_series_equal(
            df["harvest_exp_hat"],
            df["irw_need"] + df["fw_colat"] + df["fw_need"],
            rtol=1e-4,
            check_names=False,
        )
        # Column name consistent with runner.output["parameters"]
        df["disturbance_type"] = df["dist_type_name"]
        return df

    @cached_property
    def hat_extra(self) -> pandas.DataFrame:
        """Extra information from the harvest allocation tool"""
        df = self.runner.output["extras"]
        df.rename(
            columns={
                "index": "year",
                "harvest_irw_vol": "harvest_demand_irw",
                "harvest_fw_vol": "harvest_demand_fw",
            },
            inplace=True,
        )
        df["harvest_demand"] = df["harvest_demand_irw"] + df["harvest_demand_fw"]
        return df

    def expected_agg(self, groupby: Union[List[str], str]):
        """Harvest expected by the Harvest Allocation Tool (HAT) aggregated
        along grouping variables

        Get the harvest expected from disturbances allocated by hat which are
        allocated at some level of classifier groupings (other classifiers
        might have question marks i.e. where harvest can be allocated to any
        value of that particular classifier).

        In case of yearly information only, this will use extra information on pre
        determined disturbances from HAT cbm/dynamic.py.
        Use extra information from the HAT cbm/dynamic.py

        The `groupby` argument makes it possible to group on year, group on year
        and classifiers or group on the disturbance id.
        """
        if isinstance(groupby, str):
            groupby = [groupby]
        # Aggregate
        cols = ["irw_need", "fw_colat", "fw_need", "amount_exp_hat", "harvest_exp_hat"]
        df = self.hat_events.groupby(groupby)[cols].agg("sum").reset_index()
        # If grouping on years only, join demand from the economic model.
        if groupby == ["year"]:
            # msg = "Group by year. Get harvest demand and predetermined harvest "
            # msg += "information from the output extra table."
            # print(msg)
            df = df.merge(self.hat_extra, on="year", how="left")
            # Check that "harvest_exp_hat" computed from HAT disturbances is the
            # same as the sum of remaining irw and fw harvest computed at the
            # begining of cbm/dynamic.py
            # np.testing.assert_allclose(
            #     df["harvest_exp_hat"],
            #     df["remain_irw_harvest"] + df["remain_fw_harvest"],
            #     rtol=1e-4,
            # )
        return df

    @cached_property
    def provided(self):
        """Harvest provided in one country"""
        df = self.fluxes
        
        # Sum all columns that have a flux to products
        cols_to_product = df.columns[df.columns.str.contains("to_product")]
        df["to_product"] = df[cols_to_product].sum(axis=1)
        # Keep only rows with a flux to product
        selector = df.to_product > 0
        df = df[selector]
        # Check we only have 1 year since last disturbance
        time_since_last = df["time_since_last_disturbance"].unique()
        if not time_since_last == 1:
            msg = "Time since last disturbance should be one"
            msg += f"it is {time_since_last}"
            raise ValueError(msg)
        # Add wood density information by forest type
        df = df.merge(self.parent.wood_density_bark_frac, on="forest_type")

        # Convert tons of carbon to volume under bark
        df["harvest_prov_ub"] = ton_carbon_to_m3_ub(df, "to_product")
        df["harvest_prov_ob"] = ton_carbon_to_m3_ob(df, "to_product")

        # add silvicultural practices
        # Add a new column to the DataFrame
        df['silv_practice'] = None
        # Match the values in df with the keys in dist_silv_corresp
        for i in range(len(df)):
            disturbance_type = df.loc[i, 'disturbance_type']
            if disturbance_type in dist_silv_corresp:
                df.loc[i, 'silv_practice'] = dist_silv_corresp[disturbance_type]
        # #################
        # add irw fractions from input file to convert to IRW and FW
        df_irw = self.parent.irw_frac
                
        # define the scenario applicable for IRW from .yaml combo
        yaml_path = self.runner.combo.yaml_path
        with open(yaml_path, 'r') as file:
            yaml_data = file.read()
        data = yaml.safe_load(yaml_data)
        min_year = min(data['irw_frac_by_dist'].keys())
        mngm_scenario = data['irw_frac_by_dist'][min_year]
        df_irw = df_irw[df_irw['scenario'] == mngm_scenario]
        #print(self.runner.country.iso2_code)      
        
        # exclude climate which is often "?"
        # dropna to get rid of events which may not apply (e.g., deforestation)
        df= df.merge(df_irw, on = ["status","forest_type", "region",
                                    "mgmt_type","mgmt_strategy",
                                    "disturbance_type", "con_broad", 
                                    "site_index", "growth_period"], how='inner')
        
        #convert roundwood output to IRW and FW
        # add adintional split on con and broad
        df["irw_to_product_soft"] = (
            df["softwood_merch_to_product"]* df["softwood_merch_irw_frac"]+
            df["softwood_other_to_product"]*df["softwood_other_irw_frac"]+
            df["softwood_stem_snag_to_product"]*df["softwood_stem_snag_irw_frac"]+
            df["softwood_branch_snag_to_product"]*df["softwood_branch_snag_irw_frac"]
                                    )
        df["irw_to_product_hard"] = (
            df["hardwood_merch_to_product"]*df["hardwood_merch_irw_frac"]+
            df["hardwood_other_to_product"]*df["hardwood_other_irw_frac"]+
            df["hardwood_stem_snag_to_product"]*df["hardwood_stem_snag_irw_frac"]+
            df["hardwood_branch_snag_to_product"]*df["hardwood_branch_snag_irw_frac"]
                               )
        df["fw_to_product_soft"] = (
            df["softwood_merch_to_product"]*(1-df["softwood_merch_irw_frac"])+
            df["softwood_other_to_product"]*(1-df["softwood_other_irw_frac"])+
            df["softwood_stem_snag_to_product"]*(1-df["softwood_stem_snag_irw_frac"])+
            df["softwood_branch_snag_to_product"]*(1-df["softwood_branch_snag_irw_frac"])
                                )
        df["fw_to_product_hard"] = (    
            df["hardwood_merch_to_product"]*(1-df["hardwood_merch_irw_frac"])+
            df["hardwood_other_to_product"]*(1-df["hardwood_other_irw_frac"])+
            df["hardwood_stem_snag_to_product"]*(1-df["hardwood_stem_snag_irw_frac"])+
            df["hardwood_branch_snag_to_product"]*(1-df["hardwood_branch_snag_irw_frac"])
                            )
        df["irw_to_product"] = (df["irw_to_product_soft"] + df["irw_to_product_hard"] )
        df["fw_to_product"] = (df["fw_to_product_soft"] + df["fw_to_product_hard"] )
            
                              
        # Convert tons of carbon to volume under bark
        df["irw_harvest_prov_ub_con"] = ton_carbon_to_m3_ub(df, "irw_to_product_soft")
        df["irw_harvest_prov_ub_broad"] = ton_carbon_to_m3_ub(df, "irw_to_product_hard")
        df["irw_harvest_prov_ob_con"] = ton_carbon_to_m3_ob(df, "irw_to_product_soft")
        df["irw_harvest_prov_ob_broad"] = ton_carbon_to_m3_ob(df, "irw_to_product_hard")
        df["fw_harvest_prov_ub_con"] = ton_carbon_to_m3_ub(df, "fw_to_product_soft")
        df["fw_harvest_prov_ub_broad"] = ton_carbon_to_m3_ub(df, "fw_to_product_hard")
        df["fw_harvest_prov_ob_con"] = ton_carbon_to_m3_ob(df, "fw_to_product_soft")
        df["fw_harvest_prov_ob_broad"] = ton_carbon_to_m3_ob(df, "fw_to_product_hard")

        df["irw_harvest_prov_ub"] =  df["irw_harvest_prov_ub_con"] + df["irw_harvest_prov_ub_broad"]
        df["irw_harvest_prov_ob"] = df["irw_harvest_prov_ob_con"] + df["irw_harvest_prov_ob_broad"]
        df["fw_harvest_prov_ub"] = df["fw_harvest_prov_ub_con"] + df["fw_harvest_prov_ub_broad"]
        df["fw_harvest_prov_ob"] = df["fw_harvest_prov_ob_con"] + df["fw_harvest_prov_ob_broad"]
        
        # Area information
        index = ["identifier", "timestep"]
        area = self.pools[index + ["area"]]
        df = df.merge(area, on=index)
        #df.to_csv('harvest_check.csv', mode='a', index=False, header=True)
        return df

   

    @cached_property
    def provided_shares(self):
        """Harvest provided in one country on disturbance types"""
        df = self.fluxes
        
        # Sum all columns that have a flux to products
        cols_to_product = df.columns[df.columns.str.contains("to_product")]
        df["to_product"] = df[cols_to_product].sum(axis=1)
        # Keep only rows with a flux to product
        selector = df.to_product > 0
        df = df[selector]
        # Check we only have 1 year since last disturbance
        time_since_last = df["time_since_last_disturbance"].unique()
        if not time_since_last == 1:
            msg = "Time since last disturbance should be one"
            msg += f"it is {time_since_last}"
            raise ValueError(msg)
        # Add wood density information by forest type
        df = df.merge(self.parent.wood_density_bark_frac, on="forest_type")
        
        # Convert tons of carbon to volume under bark
        df["harvest_prov_ub"] = ton_carbon_to_m3_ub(df, "to_product")
        df["harvest_prov_ob"] = ton_carbon_to_m3_ob(df, "to_product")
        
        # Add a new column to the DataFrame
        df['silv_practice'] = None
                        
        # Match the values in df with the keys in dist_silv_corresp
        for i in range(len(df)):
            disturbance_type = df.loc[i, 'disturbance_type']
            if disturbance_type in dist_silv_corresp:
                df.loc[i, 'silv_practice'] = dist_silv_corresp[disturbance_type]
        
        summed_df = df.groupby(['year', 'con_broad', 'silv_practice'])['harvest_prov_ub'].sum()
        summed_df = summed_df.reset_index() 

        # Camlculate the total harvest_prov_ub for each year and con_broad
        total_harvest = summed_df.groupby(['year', 'con_broad'])['harvest_prov_ub'].transform('sum')
        # Merge the total_harvest back to the dataframe
        percentage_df = summed_df.merge(total_harvest, left_index=True, right_index=True, suffixes=('', '_total'))
        # Calculate the share of thinnings to final_cut
        percentage_df['share_to_total'] = percentage_df['harvest_prov_ub'] / percentage_df['harvest_prov_ub_total']
        return percentage_df

        # to keep this temporary omnly
        # Match the values in df with the keys in dist_silv_corresp
        #for i in range(len(df)):
        #    disturbance_type = df.loc[i, 'disturbance_type']
        #    if disturbance_type in dist_silv_corresp:
        #        df.loc[i, 'silv_practice'] = dist_silv_corresp[disturbance_type]
        #df.to_csv('overall.csv', mode='w', index=False, header=True)
        
        ##summed_df = df.groupby(['year', 'con_broad', 'silv_practice'])['harvest_prov_ub'].sum()
        ##summed_df = summed_df.reset_index() 
        
        ## Camlculate the total harvest_prov_ub for each year and con_broad
        ##total_harvest = summed_df.groupby(['year', 'con_broad'])['harvest_prov_ub'].transform('sum')
        ##total_harvest.to_csv('summed_df.csv', mode='w', index=False, header=True)
        
        ## Merge the total_harvest back to the dataframe
        ##percentage_df = summed_df.merge(total_harvest, left_index=True, right_index=True, suffixes=('', '_total'))
        ##percentage_df.to_csv('percentage_df.csv', mode='w', index=False, header=True)
        #######

        ## a) Share of harvest_prov_ub for con and broad in the sum of con and broad for each year
        #total_harvest_by_year_con_broad = df.groupby(['year', 'con_broad'])['harvest_prov_ub'].sum()
        #total_harvest_by_year = df.groupby(['year'])['harvest_prov_ub'].sum()
        #share_con_broad = total_harvest_by_year_con_broad / total_harvest_by_year
        
        ## Convert Series to DataFrame and reset index
        #share_con_broad_df = share_con_broad.reset_index(name='share_con_broad_in_total')
        
        ## b) Share of final_cut in the sum of final_cut and thinnings for each of con and broad for each year
        #final_cut_total = df[df['silv_practice'] == 'final_cut'].groupby(['year', 'con_broad'])['harvest_prov_ub'].sum()
        #total_practices = df.groupby(['year', 'con_broad'])['harvest_prov_ub'].sum()
        #share_final_cut = final_cut_total / total_practices
        
        ## Convert Series to DataFrame and reset index
        #share_final_cut_df = share_final_cut.reset_index(name='share_final_cut_in_total')
        
        #df = share_final_cut_df.merge(share_con_broad_df, on  = ['year','con_broad'])
        
        ##df.to_csv('df.csv', mode='w', index=False, header=True)  
        
        #return df
        # #################
      

    def provided_agg(self, groupby: Union[List[str], str]):
        """Aggregated version of harvest provided
        Group rows and sum all identifier rows in the same group"""
        
        # add the new columns with IRW and FW
        
        cols = (["area", "to_product", "harvest_prov_ub", "harvest_prov_ob"] +
                ["irw_to_product","fw_to_product","irw_harvest_prov_ub",
                 "irw_harvest_prov_ob", "fw_harvest_prov_ub", "fw_harvest_prov_ob", 
                 "irw_harvest_prov_ub_con","irw_harvest_prov_ub_broad", "fw_harvest_prov_ub_con", 
                 "fw_harvest_prov_ub_broad" ])
        df_agg = self.provided.groupby(groupby)[cols].agg("sum").reset_index()   
        return df_agg

    def irw_provided_2020_agg(self, groupby: Union[List[str], str]):
        
        """Aggregated version of harvest provided
        Group rows and sum all identifier rows in the same group

        >>>from eu_cbm_hat.core.continent import continent
        >>> runner = continent.combos['reference'].runners['LU'][-1]
        >>> runner.post_processor.harvest.irw_provided_2020_agg(['year', 'silv_practice'])
        
        """
        
        # add the new columns with IRW and FW
        
        cols = (["area", "to_product", "harvest_prov_ub", "harvest_prov_ob"] +
                ["irw_to_product","fw_to_product","irw_harvest_prov_ub",
                 "irw_harvest_prov_ob", "fw_harvest_prov_ub", "fw_harvest_prov_ob", 
                 "irw_harvest_prov_ub_con","irw_harvest_prov_ub_broad", "fw_harvest_prov_ub_con", 
                 "fw_harvest_prov_ub_broad", "silv_practice"])
        
        df= self.provided.groupby(groupby)[cols].agg("sum").reset_index()  

        #collect 2020
        df = df[(df['year']== 2021) | (df['year']== 2022) | (df['year']== 2023) ]
        df_agg = df.groupby(['year', 'silv_practice'])["irw_harvest_prov_ub_con","irw_harvest_prov_ub_broad"].sum()
        df_agg = df_agg.reset_index()
        df_agg['country'] = self.runner.country.country_name
        return df_agg

    def expected_provided(self, groupby: Union[List[str], str]):
        """Harvest excepted provided in one country

        There is a groupby  argument because we get the harvest expected from the
        hat output of disturbances allocated by hat which are allocated at some
        level of classifier groupings (other classifiers might have question marks
        i.e. where harvest can be allocated to any value of that particular
        classifier).

        In case the groupby argument is equal to "year", we also add the harvest
        demand from the economic model.
        """
        if isinstance(groupby, str):
            groupby = [groupby]
        # TODO: current version of harvest_exp_one_country() only contains HAT
        # disturbances. This should also provide static events that generate fluxes
        # to products especially in the historical period
        df_expected = self.expected_agg(groupby=groupby)
        df_provided = self.provided_agg(groupby=groupby)
        df = df_expected.merge(df_provided, on=groupby, how="outer")

        # Join demand from the economic model, if grouping on years only
        if groupby == ["year"]:
            # print("Group by year, adding demand columns from the economic model.")
            df = df.merge(self.demand, on=groupby)

        # Sort rows in the order of the grouping variables
        df.sort_values(groupby, inplace=True)
        return df

    @cached_property
    def disturbance_types(self):
        """Disturbance types for the purposes of harvest description,
        not suitable for natural disturbances"""
        df = self.runner.country.orig_data["disturbance_types"]
        df.rename(columns={"dist_type_name": "disturbance_type"}, inplace=True)
        df["disturbance_type"] = df["disturbance_type"].astype(int)
        df["disturbance"] = "thinning"  # Default to thinning
        dist_dict = {
            "afforestation": "afforestation",
            "deforestation|cut": "clearcut",
            "salvage": "salvage",
            "fire": "natural",
        }
        for key, value in dist_dict.items():
            selector = df["dist_desc_input"].str.contains(key, case=False)
            df.loc[selector, "disturbance"] = value
        return df

    @cached_property
    def area(self):
        """Harvest area"""
        df = self.provided
        cols = self.parent.classifiers_list + ["year"]
        cols += df.columns[df.columns.str.contains("to_product")].to_list()
        cols += ["harvest_prov_ub", "harvest_prov_ob", "area", "disturbance_type"]
        df = df[cols]
        df = df.merge(
            self.disturbance_types[["disturbance_type", "disturbance"]],
            on="disturbance_type",
        )
        return df

    def area_agg(self, groupby: Union[List[str], str]):
        """Aggregated area by classifiers or other grouping columns available"""
        if isinstance(groupby, str):
            groupby = [groupby]
        df = self.area
        cols = df.columns[df.columns.str.contains("to_product")].to_list()
        cols += ["harvest_prov_ub", "harvest_prov_ob", "area"]
        df_agg = self.area.groupby(groupby)[cols].agg("sum").reset_index()
        return df_agg




from eu_cbm_hat import eu_cbm_data_pathlib
from functools import cached_property
import pandas as pd

class HWPInput():
    """Input data for Harvested Wood Product sink computation"""

    def __init__(self, parent):
        self.parent = parent

    @cached_property
    def hwp_types(self):
        # this is the types of wood use data to be retrieved from FAOSTAT
        HWP_types = pd.read_csv(eu_cbm_data_pathlib / 'common/hwp_types.csv')
        return HWP_types
    
    @cached_property
    def eu_member_states (self):
        #list of EU MS
        EU_member_states = pd.read_csv(eu_cbm_data_pathlib / 'common/eu_member_states.csv')
        return EU_member_states
    
    @cached_property
    def faostat_bulk_data (self):
        #faostat as downloaded as bulk from FAOSTAT, namely :"Forestry_E_Europe" is a bulk download from  FAOSTAT. 
        Faostat_bulk_data = pd.read_csv(eu_cbm_data_pathlib / 'common/Forestry_E_Europe.csv', low_memory=False)
        return Faostat_bulk_data

    @cached_property
    def euwrb_stat (self):
        # input Sankey data
        EUwrb_stat = pd.read_csv(eu_cbm_data_pathlib / 'common/forestry_sankey_data.csv')
        return EUwrb_stat

    @cached_property
    def fao_rw_prod_gapfilled (self):
        ## import FAOSTAT on IRW production on con and broad, data in volume (m3), this is gapfilled for historical periods
        FAO_rw_prod_gapfilled = pd.read_csv(eu_cbm_data_pathlib / 'common/irw_con_broad_faostat.csv')
        return FAO_rw_prod_gapfilled

    @cached_property
    def crf_stat (self):
        # crf sumbissions
        CRF_stat = pd.read_csv(eu_cbm_data_pathlib / 'common/hwp_crf_submission_2023.csv')
        CRF_stat = CRF_stat.rename(columns = {'country':'area'})
        return CRF_stat

    @cached_property
    def subst_params (self):
        # substitution file
        Subst_params = pd.read_csv(eu_cbm_data_pathlib / 'common/substitution_factors.csv')
        return Subst_params

    @cached_property
    def subst_ref (self):
        # substitution reference scenario
        Subst_ref = pd.read_csv( eu_cbm_data_pathlib / 'common/substitution_reference_scenario.csv')
        return Subst_ref  
    
    @cached_property
    def silv_to_hwp (self):
        # substitution reference scenario
        Silv_to_hwp = pd.read_csv( eu_cbm_data_pathlib / 'common/silv_practices_to_hwp.csv')
        return Silv_to_hwp



"""The purpose of this script is to switch git repositories from version 1 to
version 2 of libcbm


Switch to v1 or v2:

    ipython ~/eu_cbm/eu_cbm_hat/scripts/conversion/libcbmv2/switch_git_repos.py -- --version 1

    ipython ~/eu_cbm/eu_cbm_hat/scripts/conversion/libcbmv2/switch_git_repos.py -- --version 2

"""

import argparse
import git
import sys
from eu_cbm_hat import eu_cbm_aidb_dir


parser = argparse.ArgumentParser(description='Switch git repositories for libcbm dependencies')
parser.add_argument('--version', type=str, help='Name of the version')
args = parser.parse_args()
LIBCBM_VERSION = str(args.version)


def find_sys_path(path_contains:str):
    """Find path that contains the given characters. Raise an error if there's
    not exactly one matching path. Find location of git repositories for
    libcbm_py and eu_cbm_hat
    """
    matching_paths = [path for path in sys.path if path_contains in path]
    # Exclude the path of the current script
    matching_paths = [p for p in matching_paths if "scripts" not in p]
    if len(matching_paths) != 1:
        msg = f"Expected one path containing {path_contains}, "
        msg += f"found {len(matching_paths)}\n"
        msg += f"{matching_paths}"
        raise ValueError(msg)
    return matching_paths[0]


def checkout_branch(git_repo: git.repo.base.Repo, branch_name:str):
    """Check if a repository has any changes and checkout the given branch
    """
    if git_repo.is_dirty(untracked_files=True):
        msg = f"There are changes in {git_repo}.\n"
        msg += f"Not checking out the '{branch_name}' branch."
        raise RuntimeError(msg)
    git_repo.git.checkout(branch_name)
    print(f"Checked out branch: {branch_name} of {git_repo}.")


repo_eu_cbm_hat = git.Repo(find_sys_path("eu_cbm_hat"))
repo_eu_cbm_aidb = git.Repo(eu_cbm_aidb_dir)
repo_libcbm_py = git.Repo(find_sys_path("libcbm_py"))

if LIBCBM_VERSION == "1":
    checkout_branch(repo_libcbm_py, "1.x")
    checkout_branch(repo_eu_cbm_hat, "main")
    checkout_branch(repo_eu_cbm_aidb, "main")
    assert repo_libcbm_py.active_branch.name == '1.x'
    assert repo_eu_cbm_hat.active_branch.name == 'main'

elif LIBCBM_VERSION == "2":
    checkout_branch(repo_libcbm_py, "2.x")
    checkout_branch(repo_eu_cbm_hat, "libcbm2")
    checkout_branch(repo_eu_cbm_aidb, "libcbm2")
    assert repo_libcbm_py.active_branch.name == '2.x'
    assert repo_eu_cbm_hat.active_branch.name == 'libcbm2'

else:
    raise ValueError(f"Libcbm version {LIBCBM_VERSION} not recognised")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

from typing import Union, List
from functools import cached_property
from eu_cbm_hat.post_processor.sink import Sink
from eu_cbm_hat.post_processor.harvest import Harvest
from eu_cbm_hat.post_processor.area import Area
from eu_cbm_hat.post_processor.stock import Stock
from eu_cbm_hat.post_processor.nai import NAI
from eu_cbm_hat.post_processor.growth_curve import GrowthCurve
from eu_cbm_hat.post_processor.diagnostic import Diagnostic


class PostProcessor(object):
    """
    Compute aggregates based on the pools and sink table output from the model

        >>> from eu_cbm_hat.core.continent import continent
        >>> runner = continent.combos['reference'].runners['LU'][-1]
        >>> runner.post_processor.pools
        >>> runner.post_processor.sink

        >>> runner.post_processor.pools_morf
        >>> runner.post_processor.fluxes_morf
        >>> runner.post_processor.pools_fluxes_morf

    """

    def __init__(self, parent):
        # Default attributes #
        self.parent = parent
        self.runner = parent
        self.classifiers = self.runner.output.classif_df
        self.classifiers_list = self.classifiers.columns.to_list()
        self.classifiers_list.remove("identifier")
        self.classifiers_list.remove("timestep")
        self.classifiers["year"] = self.runner.country.timestep_to_year(
            self.classifiers["timestep"]
        )
        # Index used for both self.pools_morf and self.fluxes_morf
        self.index_morf = ["year", "disturbance_type"] + self.classifiers_list
        self.state = self.runner.output["state"]
        self.params = self.runner.output["parameters"]
        # Define disturbance types
        self.afforestation_dist_type = 8
        self.deforestation_dist_type = 7
        # Check the names correspond to the one given in disturbance_types.csv
        dist_def = self.runner.country.orig_data.get_dist_description("deforestation")
        assert all(
            dist_def["dist_type_name"].head(1) == str(self.deforestation_dist_type)
        )
        dist_aff = self.runner.country.orig_data.get_dist_description("afforestation")
        assert all(
            dist_aff["dist_type_name"].head(1) == str(self.afforestation_dist_type)
        )

    def __repr__(self):
        return '%s object code "%s"' % (self.__class__, self.runner.short_name)

    def __call__(self):
        """
        xxxx.
        """
        return
        # Message #
        self.parent.log.info("Post-processing results.")
        # Lorem #
        pass


    @cached_property
    def pools(self):
        """Pools used for the sink computation

        Number of rows in the pools table:

            >>> from eu_cbm_hat.core.continent import continent
            >>> runner = continent.combos['reference'].runners['LU'][-1]
            >>> pools = runner.post_processor.pools
            >>> pools.value_counts(["year"]).reset_index()

        """
        index = ["identifier", "timestep"]
        # Data frame of pools content at the maximum disaggregated level by
        # identifier and timestep that will be sent to the other sink functions
        df = (
            self.runner.output["pools"].merge(self.classifiers, "left", on=index)
            # Add 'time_since_land_class_change' and 'time_since_last_disturbance'
            .merge(self.state, "left", on=index)
            .merge(self.params, "left", on=index)
        )
        ###################################################
        # Compute the area afforested in the current year #
        ###################################################
        # This will be used to treat afforestation soil stock change from NF.
        # This corresponds to time_since_land_class_change==1
        selector_afforest = df["status"].str.contains("AR")
        selector_afforest &= df["time_since_last_disturbance"] == 1
        selector_afforest &= df["last_disturbance_type"] == self.afforestation_dist_type
        df["area_afforested_current_year"] = df["area"] * selector_afforest
        ###################################################
        # Compute the area deforested in the current year #
        ###################################################
        selector_deforest = df["last_disturbance_type"] == self.deforestation_dist_type
        selector_deforest &= df["time_since_last_disturbance"] == 1
        df["area_deforested_current_year"] = df["area"] * selector_deforest
        return df

    @cached_property
    def fluxes(self):
        """Fluxes used for the sink computation"""
        index = ["identifier", "timestep"]
        # Data frame of fluxes at the maximum disaggregated level by
        # identifier and timestep that will be sent to the other functions
        df = (
            self.runner.output["flux"].merge(self.classifiers, "left", on=index)
            # Add 'time_since_land_class_change'
            .merge(self.state, "left", on=index)
            .merge(self.params, "left", on=index)
        )
        return df

    @cached_property
    def pools_morf(self):
        """Pools columns summed for merchantable, other, roots and foliage,
        across classifiers"""
        df = self.pools.copy()
        column_dict = {
            "merch": ['softwood_merch', 'hardwood_merch'],
            "other": ["softwood_other", "hardwood_other"],
            "roots": ["softwood_fine_roots", "hardwood_fine_roots",
                      "softwood_coarse_roots", "hardwood_coarse_roots"],
            "foliage": ["softwood_foliage", "hardwood_foliage"],
        }
        for key, cols in column_dict.items():
            df[key] = df[cols].sum(axis=1)
        selected_columns = ["area"] + list(column_dict.keys())
        df_agg = df.groupby(self.index_morf)[selected_columns].agg("sum")
        df_agg = df_agg.reset_index()
        return df_agg

    @cached_property
    def fluxes_morf(self):
        """Fluxes columns summed for merchantable to products, natural turnover
        (from merch and OWC) disturbance litter input (from merch and OWC)"""
        df = self.fluxes.copy()
        #check this df
        column_dict = {
            "merch_prod": ["softwood_merch_to_product", "hardwood_merch_to_product"],
            # I add this flux to prod
            "oth_prod": ["softwood_other_to_product", "hardwood_other_to_product"],
        }
        for key, cols in column_dict.items():
            df[key] = df[cols].sum(axis=1)
        selected_columns = list(column_dict.keys())
        selected_columns += ["turnover_merch_litter_input",
                             'turnover_oth_litter_input',
                             "disturbance_merch_litter_input",
                             'disturbance_oth_litter_input',
# I added two more flxues
                             "disturbance_merch_to_air",
                             "disturbance_oth_to_air"
                             ]
        df_agg = df.groupby(self.index_morf)[selected_columns].agg("sum")
        df_agg = df_agg.reset_index()
        return df_agg

    @cached_property
    def pools_fluxes_morf(self):
        """Merchantable pools and fluxes aggregated at the morf level on the
        columns and at the classifiers level on the rows

        To be used in the Net Annual Increment computation.
        """
        df = self.pools_morf.merge(self.fluxes_morf, on=self.index_morf)
        return df

    @cached_property
    def area(self):
        """Compute the forest carbon sink"""
        return Area(self)

    @cached_property
    def diagnostic(self):
        """Net Annual Increment"""
        return Diagnostic(self)

    @cached_property
    def growth_curve(self):
        """Growth curve"""
        return GrowthCurve(self)

    @cached_property
    def harvest(self):
        """Compute harvest expected and provided"""
        return Harvest(self)

    @cached_property
    def nai(self):
        """Net Annual Increment"""
        return NAI(self)

    @cached_property
    def sink(self):
        """Compute the forest carbon sink"""
        return Sink(self)
    
    @cached_property
    def stock(self):
        """Compute standing stocks"""
        return Stock(self)

    def sum_flux_pool(self, by: Union[List[str], str], pools: List[str]):
        """Aggregate the flux pool table over the "by" variables and for the
        given list of pools.

        Example

            >>> from eu_cbm_hat.core.continent import continent
            >>> runner_at = continent.combos["pikssp2"].runners["AT"][-1]
            >>> living_biomass_pools = [
            >>>     "softwood_merch",
            >>>     "softwood_other",
            >>>     "softwood_foliage",
            >>>     "softwood_coarse_roots",
            >>>     "softwood_fine_roots",
            >>>     "hardwood_merch",
            >>>     "hardwood_foliage",
            >>>     "hardwood_other",
            >>>     "hardwood_coarse_roots",
            >>>     "hardwood_fine_roots",
            >>> ]
            >>> runner_at.post_processor.sum_flux_pool(by="year", pools=living_biomass_pools)
            >>> runner_at.post_processor.sum_flux_pool(by=["year", "forest_type"], pools=living_biomass_pools)

        """
        df = self.runner.output.pool_flux.groupby(by)[pools].sum()
        df.reset_index(inplace=True)
        return df

    @cached_property
    def wood_density_bark_frac(self):
        """Wood density and bark fraction, ready to join

        Check wood density and bark fraction in all countries:

            >>> from eu_cbm_hat.post_processor.agg_combos import get_df_all_countries
            >>> wood_density_bark_all = get_df_all_countries(
            >>>     combo_name="reference",
            >>>     runner_method_name="post_processor.wood_density_bark_frac"
            >>> )

        """
        df = self.runner.silv.coefs.raw
        return df[["forest_type", "wood_density", "bark_frac"]].copy()
    
    @cached_property
    def irw_frac(self):
        """load irw_frac for converting output to IRW and FW, ready to join"""
        df = self.runner.silv.irw_frac.raw
        df['disturbance_type']=df['disturbance_type'].astype(int)
        
        # Get the last three columns
        cols_to_rename = df.columns[-8:]

        # Rename the columns by appending '_irw_frac'
        new_column_names = {col: f"{col}_irw_frac" for col in cols_to_rename}

        # Apply the renaming
        df.rename(columns=new_column_names, inplace=True)
        
        # Drop the 'climate' column
        df.drop(columns=['climate'], inplace=True)

        # Remove duplicate rows based on the remaining columns
        df.drop_duplicates(inplace=True)
        return df




"""Load and display growth curves for all countries"""

import pandas
from functools import cached_property

class GrowthCurve:
    """Extract information on growth curves

    Example use:

        >>> from eu_cbm_hat.core.continent import continent
        >>> runner = continent.combos['reference'].runners['LU'][-1]
        >>> runner.post_processor.growth_curve.df

    Get the growth curves for all countries

        >>> from eu_cbm_hat.post_processor.agg_combos import get_df_all_countries
        >>> growth_curve_all = get_df_all_countries(
        >>>     combo_name="reference",
        >>>     runner_method_name="post_processor.growth_curve.df"
        >>> )
 
    """

    def __init__(self, parent):
        self.parent = parent
        self.runner = parent.runner
        self.combo_name = self.runner.combo.short_name

    @cached_property
    def df_wide(self):
        """Growth curves in wide format with age classes in columns, similar to
        CBM input"""
        df = self.runner.input_data["growth_curves"]
        return df

    @cached_property
    def df(self):
        """Growth curves in long format"""
        index = self.parent.classifiers_list
        cols = self.df_wide.columns
        cols = cols[cols.str.contains("vol")]
        df = self.df_wide.melt(id_vars=index, value_vars=cols, var_name="age_class", value_name="merch_stock")
        df["age_class"] = df["age_class"].str.replace("vol", "")
        df["age_class"] = pandas.to_numeric(df["age_class"])
        return df








"""
TODO:delete this script once the refactoring is complete.
Refactoring functions to methods of a post_processor.harvest object

The purpose of this script is to compare expected and provided harvest

- Get expected harvest from the economic model
- Get provided harvest from the fluxes to products

Compute expected provided for total roundwood demand, as the sum of IRW and FW.

Usage:

    from eu_cbm_hat.core.continent import continent
    runner = continent.combos['reference'].runners['ZZ'][-1]
    runner.output["flux"]

Conversion method refactored from Viorel's Notebook at:
https://gitlab.com/bioeconomy/eu_cbm/eu_cbm_explore/-/blob/main/output_exploration/supply_vs_demand_total_volume.ipynb

"""


from typing import Union, List
import numpy as np
import pandas

from eu_cbm_hat.info.harvest import combined
from eu_cbm_hat.core.continent import continent
from eu_cbm_hat import CARBON_FRACTION_OF_BIOMASS


def ton_carbon_to_m3_ub(df, input_var):
    """Convert tons of carbon to volume in cubic meter under bark"""
    return (df[input_var] * (1 - df["bark_frac"])) / (
        CARBON_FRACTION_OF_BIOMASS * df["wood_density"]
    )


def harvest_demand(selected_scenario: str) -> pandas.DataFrame:
    """Get demand from the economic model using eu_cbm_hat/info/harvest.py

    Usage:

        >>> from eu_cbm_hat.post_processor.harvest import harvest_demand
        >>> harvest_demand("pikfair")

    """
    irw = combined["irw"]
    irw["product"] = "irw_demand"
    fw = combined["fw"]
    fw["product"] = "fw_demand"
    df = pandas.concat([irw, fw]).reset_index(drop=True)
    index = ["scenario", "iso2_code", "year"]
    df = df.pivot(index=index, columns="product", values="value").reset_index()
    df["rw_demand"] = df["fw_demand"] + df["irw_demand"]
    df = df.rename_axis(columns=None)
    return df.loc[df["scenario"] == selected_scenario]


def harvest_exp_one_country(
    combo_name: str, iso2_code: str, groupby: Union[List[str], str]
):
    """Harvest excepted in one country, as allocated by the Harvest Allocation Tool

    Get the harvest expected from the hat output of disturbances allocated by
    hat which are allocated at some level of classifier groupings (other
    classifiers might have question marks i.e. where harvest can be allocated
    to any value of that particular classifier).

    In case of yearly information only, this will use extra information on pre
    determined disturbances from HAT cbm/dynamic.py.

    The `groupby` argument makes it possible to group on year, group on year
    and classifiers or group on the disturbance id:

        >>> from eu_cbm_hat.post_processor.harvest import harvest_exp_one_country
        >>> harvest_exp_one_country("reference", "LU", "year")
        >>> harvest_exp_one_country("reference", "LU", ["year", "forest_type"])
        >>> harvest_exp_one_country("reference", "LU", ["year", "disturbance_type"])

    """
    # Load harvest expected
    runner = continent.combos[combo_name].runners[iso2_code][-1]
    events = runner.output["events"]
    events["harvest_exp_hat"] = ton_carbon_to_m3_ub(events, "amount")
    # Check that we get the same value as the sum of irw_need and fw_colat
    for col in ["harvest_exp_hat", "irw_need", "fw_colat", "fw_need"]:
        events[col] = events[col].fillna(0)
    pandas.testing.assert_series_equal(
        events["harvest_exp_hat"],
        events["irw_need"] + events["fw_colat"] + events["fw_need"],
        rtol=1e-4,
        check_names=False,
    )
    # Column name consistent with runner.output["parameters"]
    events["disturbance_type"] = events["dist_type_name"]
    # Aggregate
    cols = ["irw_need", "fw_colat", "fw_need", "amount", "harvest_exp_hat"]
    df = events.groupby(groupby)[cols].agg("sum").reset_index()
    # Rename the amount expected by the Harvest Allocation Tool
    df.rename(columns={"amount": "amount_exp_hat"}, inplace=True)

    # Join demand from the economic model, if grouping on years only
    # Use extra information from the HAT cbm/dynamic.py
    if groupby == "year" or groupby == ["year"]:
        # msg = "Group by year. Get harvest demand and predetermined harvest "
        # msg += "information from the output extra table."
        # print(msg)
        extras = runner.output["extras"].rename(columns={"index": "year"})
        df = df.merge(extras, on="year", how="left")
        # Check that "harvest_exp_hat" computed from HAT disturbances is the
        # same as the sum of remaining irw and fw harvest computed at the
        # begining of cbm/dynamic.py
        # np.testing.assert_allclose(
        #     df["harvest_exp_hat"],
        #     df["remain_irw_harvest"] + df["remain_fw_harvest"],
        #     rtol=1e-4,
        # )
        df.rename(
            columns={
                "harvest_irw_vol": "harvest_demand_irw",
                "harvest_fw_vol": "harvest_demand_fw",
            },
            inplace=True,
        )
        df["harvest_demand"] = df["harvest_demand_irw"] + df["harvest_demand_fw"]

    # Place combo name, country code and country name as first columns
    df["combo_name"] = combo_name
    df["iso2_code"] = runner.country.iso2_code
    df["country"] = runner.country.country_name
    cols = list(df.columns)
    cols = cols[-3:] + cols[:-3]
    return df[cols]


def harvest_prov_one_country(
    combo_name: str, iso2_code: str, groupby: Union[List[str], str]
):
    """Harvest provided in one country

    Usage:

        >>> from eu_cbm_hat.post_processor.harvest import harvest_prov_one_country
        >>> harvest_prov_one_country("reference", "ZZ", "year")
        >>> harvest_prov_one_country("reference", "ZZ", ["year", "forest_type"])
        >>> harvest_prov_one_country("reference", "ZZ", ["year", "disturbance_type"])

    """
    runner = continent.combos[combo_name].runners[iso2_code][-1]
    df = runner.output["flux"]
    df["year"] = runner.country.timestep_to_year(df["timestep"])
    # Merge index to be used on the output tables
    index = ["identifier", "timestep"]
    # Add classifiers
    df = df.merge(runner.output.classif_df, on=index)
    # Add wood density information by forest type
    df = df.merge(runner.silv.coefs.raw, on="forest_type")
        
    #print(df)
    # Sum all columns that have a flux to products
    cols_to_product = df.columns[df.columns.str.contains("to_product")]
    df["flux_to_product"] = df[cols_to_product].sum(axis=1)
    # Keep only rows with a flux to product
    selector = df.flux_to_product > 0
    df = df[selector]
    # Convert tons of carbon to volume under bark
    df["harvest_prov"] = ton_carbon_to_m3_ub(df, "flux_to_product")
    # Area information
    area = runner.output["pools"][index + ["area"]]
    df = df.merge(area, on=index)
    # Disturbance type information
    dist = runner.output["parameters"][index + ["disturbance_type"]]
    df = df.merge(dist, on=index)
    # Group rows and sum all identifier rows in the same group
    df_agg = (
        df.groupby(groupby)
        .agg(
            disturbed_area=("area", "sum"),
            flux_to_product=("flux_to_product", "sum"),
            harvest_prov=("harvest_prov", "sum"),
        )
        .reset_index()
    )

    # Place combo name, country code and country name as first columns
    df_agg["combo_name"] = combo_name
    df_agg["iso2_code"] = runner.country.iso2_code
    df_agg["country"] = runner.country.country_name
    cols = list(df_agg.columns)
    cols = cols[-3:] + cols[:-3]
    return df_agg[cols]


def harvest_exp_prov_one_country(
    combo_name: str, iso2_code: str, groupby: Union[List[str], str]
):
    """Harvest excepted provided in one country

    There is a groupby  argument because we get the harvest expected from the
    hat output of disturbances allocated by hat which are allocated at some
    level of classifier groupings (other classifiers might have question marks
    i.e. where harvest can be allocated to any value of that particular
    classifier).

    In case the groupby argument is equal to "year", we also add the harvest
    demand from the economic model.

    Usage:

        >>> from eu_cbm_hat.post_processor.harvest import harvest_exp_prov_one_country
        >>> import pandas
        >>> pandas.set_option('display.precision', 0) # Display rounded numbers
        >>> harvest_exp_prov_one_country("reference", "ZZ", "year")
        >>> harvest_exp_prov_one_country("reference", "ZZ", ["year", "forest_type"])
        >>> harvest_exp_prov_one_country("reference", "ZZ", ["year", "disturbance_type"])

    """
    if isinstance(groupby, str):
        groupby = [groupby]

    # TODO: current version of harvest_exp_one_country() only contains HAT
    # disturbances. This should also provide static events that generate fluxes
    # to products especially in the historical period
    df_expected = harvest_exp_one_country(
        combo_name=combo_name, iso2_code=iso2_code, groupby=groupby
    )
    df_provided = harvest_prov_one_country(
        combo_name=combo_name, iso2_code=iso2_code, groupby=groupby
    )
    #print(df_provided)
    index = ["combo_name", "iso2_code", "country"]
    index += groupby
    df = df_expected.merge(df_provided, on=index, how="outer")

    # Join demand from the economic model, if grouping on years only
    if groupby == "year":
        # print("group by year")
        harvest_scenario_name = continent.combos[combo_name].config["harvest"]
        df_demand = harvest_demand(harvest_scenario_name)
        df_demand = df_demand.loc[df_demand["iso2_code"] == iso2_code]
        index = ["iso2_code", "year"]
        df = df.merge(df_demand, on=index)

    # Sort rows in the order of the grouping variables
    df.sort_values(groupby, inplace=True)

    return df





from functools import cached_property
from typing import List, Union
import numpy as np
from eu_cbm_hat.post_processor.sink import generate_all_combinations_and_fill_na


class Area:
    """Compute the area changes through time and across classifiers

    Area grouped by status

        >>> from eu_cbm_hat.core.continent import continent
        >>> runner = continent.combos["reference"].runners["ZZ"][-1]
        >>> area_st  = runner.post_processor.area.df_agg(["year", "status"])
        >>> area_st[["year", "status", "area", "area_tm1"]].round()
            year status     area  area_tm1
        44  2021     AR     10.0       NaN
        47  2022     AR     20.0      10.0
        50  2023     AR     30.0      20.0
        53  2024     AR     40.0      30.0
        56  2025     AR     50.0      40.0
        ..   ...    ...      ...       ...
        58  2025     NF  23804.0   23130.0
        61  2026     NF  24434.0   23804.0
        64  2027     NF  25020.0   24434.0
        67  2028     NF  25563.0   25020.0
        70  2029     NF  26062.0   25563.0
        <BLANKLINE>
        [71 rows x 4 columns]

    Area grouped by status, species and age

        >>> index = ["year", "status", "forest_type", "age"]
        >>> area_st_sp_age  = runner.post_processor.area.df_agg(index)
        >>> selector = area_st_sp_age["status"] == "ForAWS"
        >>> area_st_sp_age.loc[selector, ["year", "status", "forest_type","age", "area"]]
               year  status forest_type  age       area
        702    2001  ForAWS          DF    1   14.64570
        1031   2002  ForAWS          DF    1   14.76704
        1347   2003  ForAWS          DF    1   14.97740
        1666   2004  ForAWS          DF    1    5.48501
        4280   2012  ForAWS          DF    1    5.32162
        ...     ...     ...         ...  ...        ...
        10081  2028  ForAWS          QR  188  214.30000
        10412  2029  ForAWS          QR  188  214.30000
        10082  2028  ForAWS          QR  189  214.30000
        10413  2029  ForAWS          QR  189  214.30000
        10414  2029  ForAWS          QR  190  214.30000
        <BLANKLINE>
        [10332 rows x 5 columns]

    Area grouped by status, species and age class (one every 10 years)

        >>> index = ["year", "status", "forest_type", "age_class"]
        >>> area_st_sp_agecl  = runner.post_processor.area.df_agg(index)

    Area grouped by all classifiers and age

        >>> df_agg_cl_age = runner.post_processor.area.df_agg_by_classifiers_age

    At the end of the simulation a given set of classifiers can be repeated a
    thousand times with different values of time since last disturbance, last
    disturbance type, age class etc. Area at the most level of details from the
    CBM pools output table:

        >>> df = runner.post_processor.area.df

    Total area stays constant through time

        >>> df = runner.post_processor.area.df
        >>> total_area = df.groupby("year")["area"].sum()
        >>> total_area.round().unique()
        array([111880.])

    The status changes through time

        >>> import matplotlib.pyplot as plt
        >>> area_st_wide = area_st.pivot(columns="status", index="year", values="area")
        >>> area_st_wide.plot()  # doctest: +SKIP
        >>> plt.show()  # doctest: +SKIP

    Is the area stable or changing across each of the classifiers?

        >>> cols = runner.post_processor.classifiers.columns.to_list()
        >>> cols = [x for x in cols if x not in ["identifier", "year"]]
        >>> for cl in cols:
        ...     print(cl)  # doctest: +SKIP
        ...     df_cl = df.groupby([cl, "year"])["area"].sum().reset_index()
        ...     print(df_cl["area"].round().unique())  # doctest: +SKIP

    Area changes every year through transitions due to afforestation and
    deforestation. Is the yearly area difference in ForAWS and ForNAWS equal to
    deforestation + afforestation? Group area by status and check if the
    difference in area is explained by afforestation and deforestation.

        >>> area_check = runner.post_processor.area.afforestation_deforestation(
        ...    check=True, rtol=1e-3)

    Group by "time_since..." variables

        >>> index = ["year", "status", "last_disturbance_type", 'time_since_last_disturbance']
        >>> index += ["time_since_land_class_change", "land_class"]
        >>> cols = ["area", "area_deforested_current_year", "area_afforested_current_year"]
        >>> df3 = df.query("year in [2021, 2022, 2023]").groupby(index)[cols].agg("sum")

    Group by region, climate and status

        >>> index = runner.post_processor.sink.groupby_sink
        >>> area_re_cl_st = runner.post_processor.area.df_agg(index)

    """

    def __init__(self, parent):
        self.parent = parent
        self.runner = parent.runner

    @cached_property
    def df(self):
        """Area at the most level of details available from the CBM pools
        table.

        Note: there might be many stands that have the same classifiers
        and the same age class, but a different time since last disturbance and
        a different last disturbance type.
        """
        df = self.parent.pools
        selected_cols = [
            "identifier",
            "timestep",
            "year",
            "age",
        ]
        selected_cols += self.parent.classifiers_list
        selected_cols += [
            "last_disturbance_type",
            "time_since_last_disturbance",
            "time_since_land_class_change",
            "growth_enabled",
            "enabled",
            "land_class",
            "growth_multiplier",
            "regeneration_delay",
        ]
        # Area columns
        selected_cols += df.columns[df.columns.str.contains("area")].to_list()
        # 10 year age class
        #df["age_class"] = (df["age"] / 10).round().astype(int)
        df['age_class'] = df.age // 10 + 1
        df['age_class'] = 'AGEID' + df.age_class.astype(str)

        
        return df

    @cached_property
    def df_agg_by_classifiers_age(self):
        """Area t at the classifier level and by age classes"""

        #####
        index = self.parent.classifiers_list + ["year", "age", "age_class"]
        area_columns = self.df.columns[self.df.columns.str.contains("area")].to_list()
        df_agg = self.df.groupby(index)[area_columns].agg("sum").reset_index()
        return df_agg

    def df_agg(self, groupby: Union[List[str], str] = None):
        """Area aggregated by the given grouping variables and area t-1"""
        if isinstance(groupby, str):
            groupby = [groupby]
        if "year" not in groupby:
            raise ValueError("Year has to be in the grouping variables")
        # Aggregate by the given groupby variables
        area_columns = self.df.columns[self.df.columns.str.contains("area")].to_list()
        df_agg = self.df.groupby(groupby)[area_columns].agg("sum").reset_index()
        # Index to compute the area at t-1
        time_columns = ["identifier", "year", "timestep"]
        index = [col for col in groupby if col not in time_columns]
        # Arrange by group variable with year last to prepare for shift()
        df_agg.sort_values(index + ["year"], inplace=True)
        df_agg["area_tm1"] = df_agg.groupby(index)["area"].transform(
            lambda x: x.shift()
        )
        return df_agg

    def afforestation_deforestation(self, check=True, rtol=1e-3):
        """Check afforestation and deforestation area changes recorded in
        post_processor.pools correspond to the diff in area

        If check is False do not enforce the consistency check. Set
        `check=False` for debugging purposes.

        """
        df = self.df_agg(["year", "status"])
        # To avoid NA values for AR in the middle of the time series
        df = generate_all_combinations_and_fill_na(df, ["year", "status"])
        # TODO: first year values of area_tm1 should be NA, rechange them back to NA
        df["area_diff"] = df["area"] - df["area_tm1"]
        cols = ["area_afforested_current_year", "area_deforested_current_year"]
        df1 = df.groupby("year")[cols].agg("sum").reset_index()
        df["status"] = "diff_" + df["status"]
        df2 = df.pivot(columns="status", index="year", values="area_diff").reset_index()
        df_agg = df1.merge(df2, on="year")
        if check:
            np.testing.assert_allclose(
                df_agg["area_afforested_current_year"], df_agg["diff_AR"], rtol=rtol
            )
        return df_agg
